@@ -25,11 +25,20 @@
 #include <glib.h>
 #include <seccomp.h>
 #include <sys/ioctl.h>
+#include <sys/mman.h>
+#include <unistd.h>
 #include <wtf/FileSystem.h>
+#include <wtf/UniStdExtras.h>
 #include <wtf/glib/GLibUtilities.h>
 #include <wtf/glib/GRefPtr.h>
 #include <wtf/glib/GUniquePtr.h>
 
+#if !defined(MFD_ALLOW_SEALING) && HAVE(LINUX_MEMFD_H)
+#include <linux/memfd.h>
+#endif
+
+#include "Syscalls.h"
+
 #if PLATFORM(GTK)
 #include "WaylandCompositor.h"
 #endif
@@ -40,13 +49,7 @@
 #define BASE_DIRECTORY "wpe"
 #endif
 
-#include <sys/mman.h>
-
-#ifndef MFD_ALLOW_SEALING
-
-#if HAVE(LINUX_MEMFD_H)
-
-#include <linux/memfd.h>
+#if !defined(MFD_ALLOW_SEALING) && HAVE(LINUX_MEMFD_H)
 
 // These defines were added in glibc 2.27, the same release that added memfd_create.
 // But the kernel added all of this in Linux 3.17. So it's totally safe for us to
@@ -65,9 +68,7 @@ static int memfd_create(const char* name, unsigned flags)
 {
     return syscall(__NR_memfd_create, name, flags);
 }
-#endif // #if HAVE(LINUX_MEMFD_H)
-
-#endif // #ifndef MFD_ALLOW_SEALING
+#endif // #if !defined(MFD_ALLOW_SEALING) && HAVE(LINUX_MEMFD_H)
 
 namespace WebKit {
 using namespace WebCore;
@@ -128,24 +129,55 @@ argsToFd(const Vector<CString>& args, const char *name)
     return memfd;
 }
 
-enum class DBusAddressType {
-    Normal,
-    Abstract,
-};
+static const char* applicationId(GError** error)
+{
+    GApplication* app = g_application_get_default();
+    if (!app) {
+        g_set_error_literal(error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA, "GApplication is required.");
+        return nullptr;
+    }
+
+    const char* appID = g_application_get_application_id(app);
+    if (!appID) {
+        g_set_error_literal(error, G_IO_ERROR, G_IO_ERROR_INVALID_DATA, "GApplication must have a valid ID.");
+        return nullptr;
+    }
+    return appID;
+}
+
+static int createFlatpakInfo()
+{
+    static NeverDestroyed<GUniquePtr<char>> data;
+    static size_t size;
+
+    if (!data.get()) {
+        // xdg-desktop-portal relates your name to certain permissions so we want
+        // them to be application unique which is best done via GApplication.
+        GUniqueOutPtr<GError> error;
+        const char* appID = applicationId(&error.outPtr());
+        if (!appID)
+            g_error("Unable to configure xdg-desktop-portal access in the WebKit sandbox: %s", error->message);
+
+        GUniquePtr<GKeyFile> keyFile(g_key_file_new());
+        g_key_file_set_string(keyFile.get(), "Application", "name", appID);
+        data->reset(g_key_file_to_data(keyFile.get(), &size, nullptr));
+    }
+
+    return createSealedMemFdWithData("flatpak-info", data->get(), size);
+}
 
 class XDGDBusProxyLauncher {
 public:
-    void setAddress(const char* dbusAddress, DBusAddressType addressType)
+    void setAddress(const char* dbusAddress)
     {
-        CString dbusPath = dbusAddressToPath(dbusAddress, addressType);
-        if (dbusPath.isNull())
+        if (!dbusAddress || !g_str_has_prefix(dbusAddress, "unix:"))
             return;
 
         GUniquePtr<char> appRunDir(g_build_filename(g_get_user_runtime_dir(), BASE_DIRECTORY, nullptr));
         m_proxyPath = makeProxyPath(appRunDir.get());
 
         m_socket = dbusAddress;
-        m_path = WTFMove(dbusPath);
+        m_path = dbusAddressToPath(dbusAddress);
     }
 
     bool isRunning() const { return m_isRunning; };
@@ -158,16 +190,17 @@ public:
         m_permissions = WTFMove(permissions);
     };
 
-    void launch()
+    void launch(bool enableLogging)
     {
         RELEASE_ASSERT_WITH_SECURITY_IMPLICATION(!isRunning());
 
-        if (m_socket.isNull() || m_path.isNull() || m_proxyPath.isNull())
+        if (m_socket.isNull() || m_proxyPath.isNull())
             return;
 
         int syncFds[2];
-        if (pipe2(syncFds, O_CLOEXEC) == -1)
+        if (pipe(syncFds) == -1)
             g_error("Failed to make syncfds for dbus-proxy: %s", g_strerror(errno));
+        setCloseOnExec(syncFds[0]);
 
         GUniquePtr<char> syncFdStr(g_strdup_printf("--fd=%d", syncFds[1]));
 
@@ -177,7 +210,7 @@ public:
             syncFdStr.get(),
         };
 
-        if (!g_strcmp0(g_getenv("WEBKIT_ENABLE_DBUS_PROXY_LOGGING"), "1"))
+        if (enableLogging)
             proxyArgs.append("--log");
 
         proxyArgs.appendVector(m_permissions);
@@ -185,6 +218,9 @@ public:
         int proxyFd = argsToFd(proxyArgs, "dbus-proxy");
         GUniquePtr<char> proxyArgsStr(g_strdup_printf("--args=%d", proxyFd));
 
+        // We have to run xdg-dbus-proxy under bubblewrap because we need /.flatpak-info to exist in
+        // xdg-dbus-proxy's mount namespace. Portals may use this as a trusted way to get the
+        // sandboxed process's application ID, and will break if it's missing.
         Vector<CString> args = {
             DBUS_PROXY_EXECUTABLE,
             proxyArgsStr.get(),
@@ -198,14 +234,18 @@ public:
         argv[i] = nullptr;
 
         GRefPtr<GSubprocessLauncher> launcher = adoptGRef(g_subprocess_launcher_new(G_SUBPROCESS_FLAGS_INHERIT_FDS));
-        g_subprocess_launcher_set_child_setup(launcher.get(), childSetupFunc, GINT_TO_POINTER(syncFds[1]), nullptr);
         g_subprocess_launcher_take_fd(launcher.get(), proxyFd, proxyFd);
         g_subprocess_launcher_take_fd(launcher.get(), syncFds[1], syncFds[1]);
+
         // We are purposefully leaving syncFds[0] open here.
         // xdg-dbus-proxy will exit() itself once that is closed on our exit
 
+        ProcessLauncher::LaunchOptions launchOptions;
+        launchOptions.processType = ProcessLauncher::ProcessType::DBusProxy;
+        if (!m_path.isNull())
+            launchOptions.extraSandboxPaths.add(m_path, SandboxPermission::ReadOnly);
         GUniqueOutPtr<GError> error;
-        GRefPtr<GSubprocess> process = adoptGRef(g_subprocess_launcher_spawnv(launcher.get(), argv, &error.outPtr()));
+        GRefPtr<GSubprocess> process = bubblewrapSpawn(launcher.get(), launchOptions, argv, &error.outPtr());
         if (!process.get())
             g_error("Failed to start dbus proxy: %s", error->message);
 
@@ -219,12 +259,6 @@ public:
     };
 
 private:
-    static void childSetupFunc(gpointer userdata)
-    {
-        int fd = GPOINTER_TO_INT(userdata);
-        fcntl(fd, F_SETFD, 0); // Unset CLOEXEC
-    }
-
     static CString makeProxyPath(const char* appRunDir)
     {
         if (g_mkdir_with_parents(appRunDir, 0700) == -1) {
@@ -243,30 +277,24 @@ private:
         return CString(proxySocketTemplate.get());
     };
 
-    static CString dbusAddressToPath(const char* address, DBusAddressType addressType = DBusAddressType::Normal)
+    static CString dbusAddressToPath(const char* address)
     {
-        if (!address)
-            return { };
-
-        if (!g_str_has_prefix(address, "unix:"))
-            return { };
-
-        const char* path = strstr(address, addressType == DBusAddressType::Abstract ? "abstract=" : "path=");
+        const char* path = strstr(address, "path=");
         if (!path)
             return { };
 
-        path += strlen(addressType == DBusAddressType::Abstract ? "abstract=" : "path=");
+        path += strlen("path=");
         const char* pathEnd = path;
         while (*pathEnd && *pathEnd != ',')
             pathEnd++;
 
         return CString(path, pathEnd - path);
-}
+    }
 
     CString m_socket;
     CString m_path;
     CString m_proxyPath;
-    bool m_isRunning;
+    bool m_isRunning { false };
     Vector<CString> m_permissions;
 };
 
@@ -294,15 +322,16 @@ static void bindIfExists(Vector<CString>& args, const char* path, BindFlags bind
 static void bindDBusSession(Vector<CString>& args, XDGDBusProxyLauncher& proxy)
 {
     if (!proxy.isRunning())
-        proxy.setAddress(g_getenv("DBUS_SESSION_BUS_ADDRESS"), DBusAddressType::Normal);
+        proxy.setAddress(g_getenv("DBUS_SESSION_BUS_ADDRESS"));
 
-    if (proxy.proxyPath().data()) {
+    if (!proxy.proxyPath().isNull() && !proxy.path().isNull()) {
         args.appendVector(Vector<CString>({
             "--bind", proxy.proxyPath(), proxy.path(),
         }));
     }
 }
 
+#if PLATFORM(X11)
 static void bindX11(Vector<CString>& args)
 {
     const char* display = g_getenv("DISPLAY");
@@ -325,13 +354,11 @@ static void bindX11(Vector<CString>& args)
     } else
         bindIfExists(args, xauth);
 }
+#endif
 
 #if PLATFORM(WAYLAND) && USE(EGL)
 static void bindWayland(Vector<CString>& args)
 {
-    if (PlatformDisplay::sharedDisplay().type() != PlatformDisplay::Type::Wayland)
-        return;
-
     const char* display = g_getenv("WAYLAND_DISPLAY");
     if (!display)
         display = "wayland-0";
@@ -466,7 +493,7 @@ static void bindA11y(Vector<CString>& args)
             } else {
                 GUniqueOutPtr<char> a11yAddress;
                 g_variant_get(g_dbus_message_get_body(reply.get()), "(s)", &a11yAddress.outPtr());
-                proxy.setAddress(a11yAddress.get(), DBusAddressType::Abstract);
+                proxy.setAddress(a11yAddress.get());
             }
         }
 
@@ -481,10 +508,10 @@ static void bindA11y(Vector<CString>& args)
             "--call=org.a11y.atspi.Registry=org.a11y.atspi.DeviceEventController.NotifyListenersAsync@/org/a11y/atspi/registry/deviceeventcontroller",
         });
 
-        proxy.launch();
+        proxy.launch(!g_strcmp0(g_getenv("WEBKIT_ENABLE_A11Y_DBUS_PROXY_LOGGING"), "1"));
     }
 
-    if (proxy.proxyPath().data()) {
+    if (!proxy.proxyPath().isNull()) {
         GUniquePtr<char> proxyAddress(g_strdup_printf("unix:path=%s", proxy.proxyPath().data()));
         args.appendVector(Vector<CString>({
             "--ro-bind", proxy.proxyPath(), proxy.proxyPath(),
@@ -559,6 +586,8 @@ static void bindV4l(Vector<CString>& args)
         // Not pretty but a stop-gap for pipewire anyway.
         "--dev-bind-try", "/dev/video0", "/dev/video0",
         "--dev-bind-try", "/dev/video1", "/dev/video1",
+        "--dev-bind-try", "/dev/video2", "/dev/video2",
+        "--dev-bind-try", "/dev/media0", "/dev/media0",
     }));
 }
 
@@ -573,6 +602,28 @@ static void bindSymlinksRealPath(Vector<CString>& args, const char* path)
     }
 }
 
+// Translate a libseccomp error code into an error message. libseccomp
+// mostly returns negative errno values such as -ENOMEM, but some
+// standard errno values are used for non-standard purposes where their
+// strerror() would be misleading.
+static const char* seccompStrerror(int negativeErrno)
+{
+    RELEASE_ASSERT_WITH_MESSAGE(negativeErrno < 0, "Non-negative error value from libseccomp?");
+    RELEASE_ASSERT_WITH_MESSAGE(negativeErrno > INT_MIN, "Out of range error value from libseccomp?");
+
+    switch (negativeErrno) {
+    case -EDOM:
+        return "Architecture-specific failure";
+    case -EFAULT:
+        return "Internal libseccomp failure (unknown syscall?)";
+    case -ECANCELED:
+        return "System failure beyond the control of libseccomp";
+    }
+
+    // e.g. -ENOMEM: the result of strerror() is good enough
+    return g_strerror(-negativeErrno);
+}
+
 static int setupSeccomp()
 {
     // NOTE: This is shared code (flatpak-run.c - LGPLv2.1+)
@@ -600,6 +651,10 @@ static int setupSeccomp()
     //    in common/flatpak-run.c
     //  https://git.gnome.org/browse/linux-user-chroot
     //    in src/setup-seccomp.c
+    //
+    // Other useful resources:
+    // https://github.com/systemd/systemd/blob/HEAD/src/shared/seccomp-util.c
+    // https://github.com/moby/moby/blob/HEAD/profiles/seccomp/default.json
 
 #if defined(__s390__) || defined(__s390x__) || defined(__CRIS__)
     // Architectures with CONFIG_CLONE_BACKWARDS2: the child stack
@@ -613,47 +668,70 @@ static int setupSeccomp()
     struct scmp_arg_cmp ttyArg = SCMP_A1(SCMP_CMP_MASKED_EQ, 0xFFFFFFFFu, TIOCSTI);
     struct {
         int scall;
+        int errnum;
         struct scmp_arg_cmp* arg;
     } syscallBlockList[] = {
         // Block dmesg
-        { SCMP_SYS(syslog), nullptr },
+        { SCMP_SYS(syslog), EPERM, nullptr },
         // Useless old syscall.
-        { SCMP_SYS(uselib), nullptr },
+        { SCMP_SYS(uselib), EPERM, nullptr },
         // Don't allow disabling accounting.
-        { SCMP_SYS(acct), nullptr },
+        { SCMP_SYS(acct), EPERM, nullptr },
         // 16-bit code is unnecessary in the sandbox, and modify_ldt is a
         // historic source of interesting information leaks.
-        { SCMP_SYS(modify_ldt), nullptr },
+        { SCMP_SYS(modify_ldt), EPERM, nullptr },
         // Don't allow reading current quota use.
-        { SCMP_SYS(quotactl), nullptr },
+        { SCMP_SYS(quotactl), EPERM, nullptr },
 
         // Don't allow access to the kernel keyring.
-        { SCMP_SYS(add_key), nullptr },
-        { SCMP_SYS(keyctl), nullptr },
-        { SCMP_SYS(request_key), nullptr },
+        { SCMP_SYS(add_key), EPERM, nullptr },
+        { SCMP_SYS(keyctl), EPERM, nullptr },
+        { SCMP_SYS(request_key), EPERM, nullptr },
 
         // Scary VM/NUMA ops 
-        { SCMP_SYS(move_pages), nullptr },
-        { SCMP_SYS(mbind), nullptr },
-        { SCMP_SYS(get_mempolicy), nullptr },
-        { SCMP_SYS(set_mempolicy), nullptr },
-        { SCMP_SYS(migrate_pages), nullptr },
+        { SCMP_SYS(move_pages), EPERM, nullptr },
+        { SCMP_SYS(mbind), EPERM, nullptr },
+        { SCMP_SYS(get_mempolicy), EPERM, nullptr },
+        { SCMP_SYS(set_mempolicy), EPERM, nullptr },
+        { SCMP_SYS(migrate_pages), EPERM, nullptr },
 
         // Don't allow subnamespace setups:
-        { SCMP_SYS(unshare), nullptr },
-        { SCMP_SYS(mount), nullptr },
-        { SCMP_SYS(pivot_root), nullptr },
-        { SCMP_SYS(clone), &cloneArg },
+        { SCMP_SYS(unshare), EPERM, nullptr },
+        { SCMP_SYS(setns), EPERM, nullptr },
+        { SCMP_SYS(mount), EPERM, nullptr },
+        { SCMP_SYS(umount), EPERM, nullptr },
+        { SCMP_SYS(umount2), EPERM, nullptr },
+        { SCMP_SYS(pivot_root), EPERM, nullptr },
+        { SCMP_SYS(chroot), EPERM, nullptr },
+        { SCMP_SYS(clone), EPERM, &cloneArg },
 
         // Don't allow faking input to the controlling tty (CVE-2017-5226)
-        { SCMP_SYS(ioctl), &ttyArg },
+        { SCMP_SYS(ioctl), EPERM, &ttyArg },
+
+        // seccomp can't look into clone3()'s struct clone_args to check whether
+        // the flags are OK, so we have no choice but to block clone3().
+        // Return ENOSYS so user-space will fall back to clone().
+        // (GHSA-67h7-w3jq-vh4q; see also https://github.com/moby/moby/commit/9f6b562d)
+        { SCMP_SYS(clone3), ENOSYS, nullptr },
+
+        // New mount manipulation APIs can also change our VFS. There's no
+        // legitimate reason to do these in the sandbox, so block all of them
+        // rather than thinking about which ones might be dangerous.
+        // (GHSA-67h7-w3jq-vh4q)
+        { SCMP_SYS(open_tree), ENOSYS, nullptr },
+        { SCMP_SYS(move_mount), ENOSYS, nullptr },
+        { SCMP_SYS(fsopen), ENOSYS, nullptr },
+        { SCMP_SYS(fsconfig), ENOSYS, nullptr },
+        { SCMP_SYS(fsmount), ENOSYS, nullptr },
+        { SCMP_SYS(fspick), ENOSYS, nullptr },
+        { SCMP_SYS(mount_setattr), ENOSYS, nullptr },
 
         // Profiling operations; we expect these to be done by tools from outside
         // the sandbox. In particular perf has been the source of many CVEs.
-        { SCMP_SYS(perf_event_open), nullptr },
+        { SCMP_SYS(perf_event_open), EPERM, nullptr },
         // Don't allow you to switch to bsd emulation or whatnot.
-        { SCMP_SYS(personality), nullptr },
-        { SCMP_SYS(ptrace), nullptr }
+        { SCMP_SYS(personality), EPERM, nullptr },
+        { SCMP_SYS(ptrace), EPERM, nullptr }
     };
 
     scmp_filter_ctx seccomp = seccomp_init(SCMP_ACT_ALLOW);
@@ -661,29 +739,28 @@ static int setupSeccomp()
         g_error("Failed to init seccomp");
 
     for (auto& rule : syscallBlockList) {
-        int scall = rule.scall;
         int r;
         if (rule.arg)
-            r = seccomp_rule_add(seccomp, SCMP_ACT_ERRNO(EPERM), scall, 1, *rule.arg);
+            r = seccomp_rule_add(seccomp, SCMP_ACT_ERRNO(rule.errnum), rule.scall, 1, *rule.arg);
         else
-            r = seccomp_rule_add(seccomp, SCMP_ACT_ERRNO(EPERM), scall, 0);
-        if (r == -EFAULT) {
-            seccomp_release(seccomp);
-            g_error("Failed to add seccomp rule");
-        }
+            r = seccomp_rule_add(seccomp, SCMP_ACT_ERRNO(rule.errnum), rule.scall, 0);
+        // EFAULT means "internal libseccomp error", but in practice we get
+        // this for syscall numbers added via Syscalls.h (flatpak-syscalls-private.h)
+        // when trying to filter them on a non-native architecture, because
+        // libseccomp cannot map the syscall number to a name and back to a
+        // number for the non-native architecture.
+        if (r == -EFAULT)
+            g_info("Unable to block syscall %d: syscall not known to libseccomp?", rule.scall);
+        else if (r < 0)
+            g_error("Failed to block syscall %d: %s", rule.scall, seccompStrerror(r));
     }
 
     int tmpfd = memfd_create("seccomp-bpf", 0);
-    if (tmpfd == -1) {
-        seccomp_release(seccomp);
+    if (tmpfd == -1)
         g_error("Failed to create memfd: %s", g_strerror(errno));
-    }
 
-    if (seccomp_export_bpf(seccomp, tmpfd)) {
-        seccomp_release(seccomp);
-        close(tmpfd);
-        g_error("Failed to export seccomp bpf");
-    }
+    if (int r = seccomp_export_bpf(seccomp, tmpfd))
+        g_error("Failed to export seccomp bpf: %s", seccompStrerror(r));
 
     if (lseek(tmpfd, 0, SEEK_SET) < 0)
         g_error("lseek failed: %s", g_strerror(errno));
@@ -692,52 +769,54 @@ static int setupSeccomp()
     return tmpfd;
 }
 
-static int createFlatpakInfo()
+static bool shouldUnshareNetwork(ProcessLauncher::ProcessType processType)
 {
-    GUniquePtr<GKeyFile> keyFile(g_key_file_new());
+    // xdg-dbus-proxy needs access to host abstract sockets to connect to the a11y bus. Secure
+    // host services must not use abstract sockets.
+    if (processType == ProcessLauncher::ProcessType::DBusProxy)
+        return false;
 
-    // xdg-desktop-portal relates your name to certain permissions so we want
-    // them to be application unique which is best done via GApplication.
-    GApplication* app = g_application_get_default();
-    if (!app) {
-        g_warning("GApplication is required for xdg-desktop-portal access in the WebKit sandbox. Actions that require xdg-desktop-portal will be broken.");
-        return -1;
+#if PLATFORM(X11)
+    // Also, the web process needs access to host networking if the X server is running over TCP or
+    // on a different host's Unix socket; this is likely the case if the first character of DISPLAY
+    // is not a colon.
+    if (processType == ProcessLauncher::ProcessType::Web && PlatformDisplay::sharedDisplay().type() == PlatformDisplay::Type::X11) {
+        const char* display = g_getenv("DISPLAY");
+        if (display && display[0] != ':')
+            return false;
     }
-    g_key_file_set_string(keyFile.get(), "Application", "name", g_application_get_application_id(app));
+#endif
 
-    size_t size;
-    GUniqueOutPtr<GError> error;
-    GUniquePtr<char> data(g_key_file_to_data(keyFile.get(), &size, &error.outPtr()));
-    if (error.get()) {
-        g_warning("%s", error->message);
-        return -1;
+    // Otherwise, only the network process should have network access. If we are the network
+    // process, then we are not sandboxed and have already bailed out before this point.
+    return true;
+}
+
+static void addExtraPaths(const HashMap<CString, SandboxPermission>& paths, Vector<CString>& args)
+{
+    for (const auto& pathAndPermission : paths) {
+        args.appendVector(Vector<CString>({
+            pathAndPermission.value == SandboxPermission::ReadOnly ? "--ro-bind-try": "--bind-try",
+            pathAndPermission.key, pathAndPermission.key
+        }));
     }
-
-    return createSealedMemFdWithData("flatpak-info", data.get(), size);
 }
 
 GRefPtr<GSubprocess> bubblewrapSpawn(GSubprocessLauncher* launcher, const ProcessLauncher::LaunchOptions& launchOptions, char** argv, GError **error)
 {
     ASSERT(launcher);
 
-#if ENABLE(NETSCAPE_PLUGIN_API)
-    // It is impossible to know what access arbitrary plugins need and since it is for legacy
-    // reasons lets just leave it unsandboxed.
-    if (launchOptions.processType == ProcessLauncher::ProcessType::Plugin)
-        return adoptGRef(g_subprocess_launcher_spawnv(launcher, argv, error));
-#endif
-
     // For now we are just considering the network process trusted as it
     // requires a lot of access but doesn't execute arbitrary code like
     // the WebProcess where our focus lies.
     if (launchOptions.processType == ProcessLauncher::ProcessType::Network)
         return adoptGRef(g_subprocess_launcher_spawnv(launcher, argv, error));
 
+    const char* runDir = g_get_user_runtime_dir();
     Vector<CString> sandboxArgs = {
         "--die-with-parent",
         "--unshare-pid",
         "--unshare-uts",
-        "--unshare-net",
 
         // We assume /etc has safe permissions.
         // At a later point we can start masking privacy-concerning files.
@@ -746,7 +825,8 @@ GRefPtr<GSubprocess> bubblewrapSpawn(GSubprocessLauncher* launcher, const Proces
         "--proc", "/proc",
         "--tmpfs", "/tmp",
         "--unsetenv", "TMPDIR",
-        "--dir", "/run",
+        "--dir", runDir,
+        "--setenv", "XDG_RUNTIME_DIR", runDir,
         "--symlink", "../run", "/var/run",
         "--symlink", "../tmp", "/var/tmp",
         "--ro-bind", "/sys/block", "/sys/block",
@@ -765,12 +845,34 @@ GRefPtr<GSubprocess> bubblewrapSpawn(GSubprocessLauncher* launcher, const Proces
         "--ro-bind-try", "/usr/lib", "/usr/lib",
         "--ro-bind-try", "/usr/local/lib", "/usr/local/lib",
         "--ro-bind-try", LIBDIR, LIBDIR,
+#if CPU(ADDRESS64)
         "--ro-bind-try", "/lib64", "/lib64",
         "--ro-bind-try", "/usr/lib64", "/usr/lib64",
         "--ro-bind-try", "/usr/local/lib64", "/usr/local/lib64",
+#else
+        "--ro-bind-try", "/lib32", "/lib32",
+        "--ro-bind-try", "/usr/lib32", "/usr/lib32",
+        "--ro-bind-try", "/usr/local/lib32", "/usr/local/lib32",
+#endif
 
         "--ro-bind-try", PKGLIBEXECDIR, PKGLIBEXECDIR,
     };
+
+    if (launchOptions.processType == ProcessLauncher::ProcessType::DBusProxy) {
+        sandboxArgs.appendVector(Vector<CString>({
+            "--ro-bind", DBUS_PROXY_EXECUTABLE, DBUS_PROXY_EXECUTABLE,
+            // This is a lot of access, but xdg-dbus-proxy is trusted so that's OK. It's sandboxed
+            // only because we have to mount .flatpak-info in its mount namespace. The user rundir
+            // is where we mount our proxy socket.
+            "--bind", runDir, runDir,
+        }));
+
+        addExtraPaths(launchOptions.extraSandboxPaths, sandboxArgs);
+    }
+
+    if (shouldUnshareNetwork(launchOptions.processType))
+        sandboxArgs.append("--unshare-net");
+
     // We would have to parse ld config files for more info.
     bindPathVar(sandboxArgs, "LD_LIBRARY_PATH");
 
@@ -803,21 +905,18 @@ GRefPtr<GSubprocess> bubblewrapSpawn(GSubprocessLauncher* launcher, const Proces
     if (launchOptions.processType == ProcessLauncher::ProcessType::Web) {
         static XDGDBusProxyLauncher proxy;
 
-        // If Wayland in use don't grant X11
 #if PLATFORM(WAYLAND) && USE(EGL)
         if (PlatformDisplay::sharedDisplay().type() == PlatformDisplay::Type::Wayland) {
             bindWayland(sandboxArgs);
             sandboxArgs.append("--unshare-ipc");
-        } else
-#endif
-            bindX11(sandboxArgs);
-
-        for (const auto& pathAndPermission : launchOptions.extraWebProcessSandboxPaths) {
-            sandboxArgs.appendVector(Vector<CString>({
-                pathAndPermission.value == SandboxPermission::ReadOnly ? "--ro-bind-try": "--bind-try",
-                pathAndPermission.key, pathAndPermission.key
-            }));
         }
+#endif
+#if PLATFORM(X11)
+        if (PlatformDisplay::sharedDisplay().type() == PlatformDisplay::Type::X11)
+            bindX11(sandboxArgs);
+#endif
+
+        addExtraPaths(launchOptions.extraSandboxPaths, sandboxArgs);
 
         Vector<String> extraPaths = { "applicationCacheDirectory", "mediaKeysDirectory", "waylandSocket", "webSQLDatabaseDirectory" };
         for (const auto& path : extraPaths) {
@@ -850,7 +949,7 @@ GRefPtr<GSubprocess> bubblewrapSpawn(GSubprocessLauncher* launcher, const Proces
                 permissions.append("--talk=org.freedesktop.portal.Desktop");
             }
             proxy.setPermissions(WTFMove(permissions));
-            proxy.launch();
+            proxy.launch(!g_strcmp0(g_getenv("WEBKIT_ENABLE_DBUS_PROXY_LOGGING"), "1"));
         }
     } else {
         // Only X11 users need this for XShm which is only the Web process.
@@ -860,14 +959,14 @@ GRefPtr<GSubprocess> bubblewrapSpawn(GSubprocessLauncher* launcher, const Proces
 #if ENABLE(DEVELOPER_MODE)
     const char* execDirectory = g_getenv("WEBKIT_EXEC_PATH");
     if (execDirectory) {
-        String parentDir = FileSystem::directoryName(FileSystem::stringFromFileSystemRepresentation(execDirectory));
+        String parentDir = FileSystem::parentPath(FileSystem::stringFromFileSystemRepresentation(execDirectory));
         bindIfExists(sandboxArgs, parentDir.utf8().data());
     }
 
     CString executablePath = getCurrentExecutablePath();
     if (!executablePath.isNull()) {
         // Our executable is `/foo/bar/bin/Process`, we want `/foo/bar` as a usable prefix
-        String parentDir = FileSystem::directoryName(FileSystem::directoryName(FileSystem::stringFromFileSystemRepresentation(executablePath.data())));
+        String parentDir = FileSystem::parentPath(FileSystem::parentPath(FileSystem::stringFromFileSystemRepresentation(executablePath.data())));
         bindIfExists(sandboxArgs, parentDir.utf8().data());
     }
 #endif
